;;; cocoa.scm

(declare
  (export ns:application-main ns:beep ns:log ns:rect-fill cocoa:run
          ns:rect->locative ns:point->locative
          ns:size->locative ns:range->locative) )

(require 'objc-support)
(require-for-syntax 'srfi-13) ;; for string-upcase

#>
#import <Cocoa/Cocoa.h>
<#

;;; Wrappers for global functions

;; http://developer.apple.com/documentation/Cocoa/Reference/ApplicationKit/ObjC_classic/Functions/AppKitFunctions.html

(define (ns:application-main)
  (define NSApplicationMain
    (foreign-safe-lambda int "NSApplicationMain" int c-pointer))
  (apply NSApplicationMain (receive (argc+argv))))

(define ns:beep (foreign-lambda void "NSBeep"))

(define (ns:log fstr . args)
  (define NSLog (foreign-lambda* void ((c-string s))
                 "NSString *str = [[NSString alloc] initWithUTF8String: s];"
		 "NSLog(str);"
		 "[str release];"))
  (NSLog (apply sprintf fstr args)))
  
(define (ns:rect-fill r)
  (define NSRectFill (foreign-lambda* void (((pointer "NSRect") r))
                                      "NSRectFill(*r);"))
  (NSRectFill (ns:rect->locative r)))

;;; Wrappers for enums

;; http://developer.apple.com/documentation/Cocoa/Reference/ApplicationKit/ObjC_classic/TypesAndConstants/AppKitTypes.html

;; None yet.

;;; Debugging

;; Resume the current application.
(define-objc-classes (cocoa:NSApplication NSApplication))
(define (cocoa:run)
    @[ @[cocoa:NSApplication sharedApplication] run])


;;; Conversion of ns:struct types into "locatives" that may be passed to functions
;;; that require C pointers.  The locative points to a *new* byte vector
;;; containing a C representation of the type -- not the ns:struct record itself.

;; Note: these are really intended for passing structs by value to global (non-method)
;; functions.  Passing by reference to a function will work, but any modifications
;; to the struct will be discarded.

;; Transformation:
;; (define-ns:struct->locative rect)  =>
;; (define ns:rect->locative
;;   (let ((size (objc:sizeof-type objc:NSRECT)))
;;     (lambda (r)
;;       (ns:rect->ref r (make-locative (make-byte-vector size))))))

(define-macro (define-ns:struct->locative type)
  (define (symbol-conc . symbols)
    (string->symbol (apply conc symbols)))

  (let ((struct-name (conc "ns:" type))
        (type-name   (conc "objc:NS"
                           (string-upcase (->string type)))))
  `(define ,(symbol-conc struct-name '->locative)
     (let ((size (objc:sizeof-type ,(string->symbol type-name))))
       (lambda (r)
         (,(symbol-conc struct-name '->ref)
          r (make-locative (make-byte-vector size))))))))

;;;; The ns:struct->locative conversions

(define-ns:struct->locative rect)
(define-ns:struct->locative point)
(define-ns:struct->locative size)
(define-ns:struct->locative range)

